from .lib.auth0 import get_access_token


class TokenManager(object):

    def __init__(self, settings, storage):
        self.storage = storage
        self.settings = settings

    def get_token(self, service_name):
        token = self.storage.get('token:{}'.format(service_name))
        return token or self.refresh_token(service_name)

    def refresh_token(self, service_name):
        token = get_access_token(
            self.settings['api_ids'][service_name],
            self.settings['client_id'],
            self.settings['client_secret'],
            self.settings['domain'],
        )

        self.storage.set(
            'token:{}'.format(service_name),
            token
        )
        return token

    def refresh_all_tokens(self):
        for service_name in self.settings['api_ids']:
            self.refresh_token(service_name)



import requests


def get_access_token(audience, client_id, client_secret, domain):
    response = requests.post(
        'https://{}/oauth/token'.format(domain),
        json={
            'grant_type': 'client_credentials',
            'client_id': client_id,
            'client_secret': client_secret,
            'audience': audience,
        }
    )
    response.raise_for_status()
    return response.json().get('access_token')



from unittest import TestCase
from unittest.mock import Mock, patch

from mbq.client.client import ServiceClient


class CorrelationIDTestCase(TestCase):

    def test_correlation_id_getter_provided(self):
        self.client = ServiceClient(
            'https://foo.com/',
            correlation_id_getter=lambda: 'hello-world'
        )
        with patch('requests.get') as requests_mock:
            self.client.get('/url')
            self.assertEqual(1, requests_mock.call_count)
            headers = requests_mock.call_args[1]['headers']
            self.assertEqual('hello-world', headers['X-Correlation-Id'])

    def test_correlation_id_getter_not_provided(self):
        self.client = ServiceClient('https://foo.com/')
        with patch('requests.get') as requests_mock:
            self.client.get('/url')
            self.assertEqual(1, requests_mock.call_count)
            headers = requests_mock.call_args[1]['headers']
            self.assertIsNone(headers.get('X-Correlation-Id'))

    def test_multivalued_correlation_id_getter_provided(self):
        cid_getter_mock = Mock()
        cid_getter_mock.side_effect = ['hello-world', 'goodbye-world']
        self.client = ServiceClient(
            'https://foo.com/',
            correlation_id_getter=cid_getter_mock
        )

        with patch('requests.get') as requests_mock:
            self.client.get('/url')
            self.assertEqual(1, requests_mock.call_count)
            headers = requests_mock.call_args[1]['headers']
            self.assertEqual('hello-world', headers.get('X-Correlation-Id'))

            self.client.get('/url')
            self.assertEqual(2, requests_mock.call_count)
            headers = requests_mock.call_args[1]['headers']
            self.assertEqual('goodbye-world', headers.get('X-Correlation-Id'))



import os
from tempfile import NamedTemporaryFile
from unittest import TestCase

from mbq.client.storage import FileStorage


class FileStorageTestCase(TestCase):

    def setUp(self):
        self.test_filename = NamedTemporaryFile(delete=False).name
        self.storage = FileStorage(self.test_filename)

    def tearDown(self):
        os.remove(self.test_filename)

    def test_storage(self):
        # When the file is empty, we should receive None for any key.
        self.assertIsNone(self.storage.get('key1'))

        # We should be able to write a key/value,
        self.storage.set('key1', 'value1')
        # retrieve it,
        self.assertEqual(self.storage.get('key1'), 'value1')
        # and still receive None for missing keys.
        self.assertIsNone(self.storage.get('key2'))

        # We should be able to write a 2nd key,
        self.storage.set('key2', 'value2')
        # retrieve it,
        self.assertEqual(self.storage.get('key2'), 'value2')
        # still retrieve the earlier key we wrote,
        self.assertEqual(self.storage.get('key1'), 'value1')
        # and still receive None for missing keys.
        self.assertIsNone(self.storage.get('key3'))

        # We should be able to update an existing key,
        self.storage.set('key2', 'some-new-value')
        # see the value change when retrieving,
        self.assertEqual(self.storage.get('key2'), 'some-new-value')
        # the other values should remain unchanged,
        self.assertEqual(self.storage.get('key1'), 'value1')
        # and we should still receive None for missing keys.
        self.assertIsNone(self.storage.get('key3'))

        # If we re-init the storage object with the same file,
        self.storage = FileStorage(self.test_filename)
        # all keys should be persisted.
        self.assertEqual(self.storage.get('key2'), 'some-new-value')
        self.assertEqual(self.storage.get('key1'), 'value1')
        self.assertIsNone(self.storage.get('key3'))



from .authenticator import Authenticator  # noqa
from .client import ServiceClient  # noqa
from .storage import DjangoCacheStorage, FileStorage  # noqa
from .token_manager import TokenManager  # noqa



from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from mbq.client.token_manager import TokenManager


class TokenManagerTestCase(TestCase):

    def setUp(self):

        self.get = MagicMock()

        self.set = MagicMock()

        self.storage_mock = MagicMock(
            set=self.set,
            get=self.get
        )

        self.get_auth0_token = MagicMock()

        get_auth0_token_patch = patch(
            'mbq.client.token_manager.get_access_token',
            self.get_auth0_token
        )
        get_auth0_token_patch.start()
        self.addCleanup(get_auth0_token_patch.stop)

        self.token_manager = TokenManager(
            {
                'api_ids': {'test': 'test_id', 'other_service': 'other_id'},
                'client_id': 'client_id',
                'client_secret': 'shh... it\'s a secret',
                'domain': 'auth0.com',
            },
            storage=self.storage_mock
        )

    def test_refresh_token(self):

        expected_token = 'fake_token'
        self.get_auth0_token.return_value = expected_token

        token = self.token_manager.refresh_token('test')

        self.assertEqual(token, expected_token)

        self.get_auth0_token.assert_called_once_with(
            'test_id',
            'client_id',
            'shh... it\'s a secret',
            'auth0.com'
        )

        self.set.assert_called_once_with(
            'token:test',
            expected_token
        )

    def test_refresh_all_tokens(self):

        refresh_token_mock = MagicMock()
        self.token_manager.refresh_token = refresh_token_mock

        self.token_manager.refresh_all_tokens()

        refresh_token_mock.assert_has_calls(
            [call('test'), call('other_service')], any_order=True
        )



import logging
from io import BufferedReader, BytesIO
from urllib.parse import urlparse

import requests


logger = logging.getLogger(__name__)


class ServiceClient:
    def __init__(self, api_url, auth=None, headers=None, post_process_response=None,
                 correlation_id_getter=None, default_timeout=30):
        """ We use correlation_ids to track the flow of a request between different services.
        It is the responsibility of users of ServiceClient to implement a correlation_id_getter
        """
        self._api_url = api_url
        self._auth = auth
        self._headers = headers
        self._post_process_response = post_process_response
        self._timeout = default_timeout
        self.correlation_id_getter = correlation_id_getter

    def _make_url(self, url):
        parsed = urlparse(url)
        if not any([parsed.scheme, parsed.netloc]):
            url = '{}{}'.format(self._api_url, url)

        return url

    def _make_request(self, method, url, *args, **kwargs):
        if self._headers:
            kwargs['headers'] = dict(self._headers, **kwargs.get('headers', {}))
        else:
            kwargs['headers'] = {}

        if self.correlation_id_getter is not None:
            cid = self.correlation_id_getter()
            if cid is not None:
                kwargs['headers']['X-Correlation-Id'] = cid

        if self._auth and 'auth' not in kwargs:
            kwargs['auth'] = self._auth

        if self._timeout and 'timeout' not in kwargs:
            kwargs['timeout'] = self._timeout

        url = self._make_url(url)

        response = getattr(requests, method)(url, *args, **kwargs)

        try:
            response.raise_for_status()
        except: # noqa
            msg = '{} Bad Response: {}'.format(response.status_code, response.content)
            logger.error(msg)
            raise

        try:
            if self._post_process_response is not None:
                data = self._post_process_response(response.json())
            else:
                data = response.json()
        except: # noqa
            if response.headers.get('content-length') == '0':
                data = None
            else:
                data = BufferedReader(BytesIO(response.content))

        return data

    def get(self, *args, **kwargs):
        return self._make_request('get', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self._make_request('post', *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self._make_request('patch', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._make_request('put', *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._make_request('delete', *args, **kwargs)



import json


class DjangoCacheStorage(object):

    def __init__(self, cache, timeout=None):
        self.cache = cache
        self.timeout = timeout

    def set(self, key, value):
        self.cache.set(key, value, timeout=self.timeout)

    def get(self, key):
        return self.cache.get(key)


class FileStorage(object):

    def __init__(self, filename):
        self.filename = filename
        open(self.filename, 'a').close()

    def _get_tokens_from_file(self):
        with open(self.filename, 'r') as f:
            try:
                return json.load(f)
            except ValueError:
                return {}

    def set(self, key, value):
        tokens = self._get_tokens_from_file()

        tokens[key] = value

        with open(self.filename, 'w') as f:
            json.dump(tokens, f)

    def get(self, key):
        return self._get_tokens_from_file().get(key)



from typing import Dict
from unittest import TestCase
from unittest.mock import Mock

from .. import permissions as sut


class TestOSCoreClient:
    def __init__(self, data):
        self.data = data

    def fetch_permissions(
        self, person_id: str, org_ref: sut.UUIDType
    ) -> sut.FetchedPermissionsDoc:
        result: Dict = {"global": self.data.get("global", [])}
        result.update({org_ref: self.data.get(org_ref, [])})
        return result

    def fetch_permissions_for_location(
        self, person_id: str, location_id: int, location_type: sut.RefType
    ) -> sut.FetchedPermissionsDoc:
        result: Dict = {"global": self.data.get("global", [])}
        key = f"{location_type}:{location_id}"
        result.update({key: self.data.get(key, [])})
        return result


class PermissionsClientTest(TestCase):
    def setUp(self):
        test_data = {
            "org": ["read:invoices"],
            "company:1": ["read:orders", "write:orders"],
            "vendor:2": ["read:team"],
            "global": ["read:global"],
        }
        self.client = sut.PermissionsClient(
            TestOSCoreClient(test_data), cache_name=None
        )

    def test_cache_key(self):
        self.assertEqual(
            self.client._cache_key("person", "123"), "permissions_client:person:123"
        )
        self.assertEqual(
            self.client._cache_key("person2", "456", "company"),
            "permissions_client:person2:456:company",
        )

    def test_global_cache_key(self):
        self.assertEqual(
            self.client._global_cache_key("person"), "permissions_client:person:global"
        )

    def test_cache_read_no_cache(self):
        # Just testing that this doesn't error
        self.assertEqual(self.client._cache_read("person", "org"), None)

    def test_cache_read_with_empty_cache(self):
        cache_mock = Mock()
        self.client.cache = cache_mock

        cache_mock.get_many.return_value = {}

        self.assertEqual(self.client._cache_read("person", "org"), None)

        cache_mock.get_many.assert_called_once_with(
            ["permissions_client:person:global", "permissions_client:person:org"]
        )

    def test_cache_read_with_populated_cache(self):
        cache_mock = Mock()
        self.client.cache = cache_mock

        cached_doc = {
            "permissions_client:person:global": "|",
            "permissions_client:person:org": "read:invoices|",
        }
        cache_mock.get_many.return_value = cached_doc

        self.assertEqual(self.client._cache_read("person", "org"), cached_doc)

        cache_mock.get_many.assert_called_once_with(
            ["permissions_client:person:global", "permissions_client:person:org"]
        )

    def test_cache_transform(self):
        permissions_doc = {"org": ["read:invoices"], "global": []}

        cached_doc = {
            "permissions_client:person:org": "read:invoices|",
            "permissions_client:person:global": "|",
        }

        self.assertEqual(
            self.client._cache_transform("person", permissions_doc), cached_doc
        )

    def test_cache_write(self):
        cache_mock = Mock()
        self.client.cache = cache_mock
        self.client.cache_period_seconds = 123

        permissions_doc = {"org": ["read:invoices"], "global": []}

        cached_doc = {
            "permissions_client:person:org": "read:invoices|",
            "permissions_client:person:global": "|",
        }

        self.assertEqual(
            self.client._cache_transform("person", permissions_doc), cached_doc
        )

        self.client._cache_write(cached_doc)
        cache_mock.set_many.assert_called_once_with(cached_doc, timeout=123)

    def test_has_permission_org(self):
        self.assertTrue(self.client.has_permission("person", "read:invoices", "org"))
        self.assertFalse(self.client.has_permission("person", "read:stuff", "org"))
        # Implicit global
        self.assertTrue(self.client.has_permission("person", "read:global", "org"))

    def test_has_permission_location(self):
        self.assertTrue(
            self.client.has_permission("person", "read:orders", 1, "company")
        )
        self.assertFalse(
            self.client.has_permission("person", "read:stuff", 1, "company")
        )
        self.assertTrue(self.client.has_permission("person", "read:team", 2, "vendor"))
        self.assertFalse(
            self.client.has_permission("person", "read:stuff", 2, "vendor")
        )
        # Implicit global
        self.assertTrue(
            self.client.has_permission("person", "read:global", 1, "company")
        )
        self.assertTrue(
            self.client.has_permission("person", "read:global", 2, "vendor")
        )

    def test_has_global_permission(self):
        self.assertTrue(self.client.has_global_permission("person", "read:global"))
        self.assertFalse(self.client.has_global_permission("person", "read:stuff"))



__author__ = 'Managed by Q, Inc.'
__author_email__ = 'open-source@managedbyq.com'
__description__ = 'MBQ Service Client'
__license__ = 'Apache 2.0'
__title__ = 'mbq.client'
__url__ = 'https://github.com/managedbyq/mbq.client'
__version__ = '1.1.0'



import logging
import urllib
from copy import copy
from typing import Dict, List, Optional, Union, cast, overload
from uuid import UUID

import requests
from typing_extensions import Literal, Protocol

from .. import ServiceClient


logger = logging.getLogger(__name__)

UUIDType = Union[str, UUID]
# External type returned from internal and external OS Core clients. Keys
# are the org refs if UUIDs, or {ref_type}:{ref_id} if legacy int types.
# Values are lists of scopes. Should also include a "global" literal key.
FetchedPermissionsDoc = Dict[str, List[str]]
# Internal type stored in the cache. Keys are cache keys with prefixes, colons, etc.
# Values are pipe-delimited strings with an additional pipe on the end.
CachedPermissionsDoc = Dict[str, str]
RefType = Union[Literal["company", "vendor"]]


class ClientError(Exception):
    """Raised from within OSCoreClient implementations to denote the fetch
    failed due to a client-side error.
    """

    pass


class ServerError(Exception):
    """Raised from within OSCoreClient implementations to denote the fetch
    failed due to a server-side error.
    """

    pass


class OSCoreClient(Protocol):
    def fetch_permissions(
        self, person_id: str, org_ref: UUIDType
    ) -> FetchedPermissionsDoc:
        ...

    def fetch_permissions_for_location(
        self, person_id: str, location_id: int, location_type: RefType
    ) -> FetchedPermissionsDoc:
        ...


class OSCoreServiceClient:
    def __init__(self, client: ServiceClient):
        # The copying and munging here is attempting to deal with differences
        # in how the individual ServiceClients are configured. We can get rid
        # of it if we standardize within the services.

        # Note that we are doing a shallow copy so the Authenticator instance
        # will be shared
        self.client = copy(client)

        self.client._post_process_response = None
        self.client._headers = None

        parsed = urllib.parse.urlparse(self.client._api_url)
        self.client._api_url = f"{parsed.scheme}://{parsed.netloc}"

    def fetch_permissions(
        self, person_id: str, org_ref: UUIDType
    ) -> FetchedPermissionsDoc:
        logger.debug(f"Fetching permissions from OS Core: {person_id}, {org_ref}")

        try:
            return self.client.get(
                f"/api/v1/people/{person_id}/permissions", params={"org_ref": org_ref}
            )["objects"]
        except requests.exceptions.HTTPError as e:
            response = getattr(e, "response", None)
            if response and response.status_code // 100 == 4:
                raise ClientError("Invalid request") from e
            raise ServerError("Server error") from e
        except Exception as e:
            raise ServerError("Server error") from e

    def fetch_permissions_for_location(
        self, person_id: str, location_id: int, location_type: RefType
    ) -> FetchedPermissionsDoc:
        logger.debug(
            f"Fetching permissions from OS Core: {person_id}, {location_type} {location_id}"
        )

        try:
            return self.client.get(
                f"/api/v1/people/{person_id}/permissions",
                params={"location_id": location_id, "location_type": location_type},
            )["objects"]
        except requests.exceptions.HTTPError as e:
            response = getattr(e, "response", None)
            if response and response.status_code // 100 == 4:
                raise ClientError("Invalid request") from e
            raise ServerError("Server error") from e
        except Exception as e:
            raise ServerError("Server error") from e


class PermissionsClient:
    """Cache-aware client for consuming the Permissions API from OS Core.

    os_core_client: OSCoreClient Protocol implementation used to talk to OS Core. From
                    remote services, this should be a wrapped ServiceClient instance.
                    Use the provided OSCoreServiceClient wrapper in this contrib.
                    From OS Core itself, this should make local function calls.
    cache_name: Name of the Django cache to use, default "default". Pass None
                to disable caching.
    cache_period_seconds: Expiration time on cache keys in seconds.
    """

    _cache_prefix = "permissions_client"

    def __init__(
        self,
        os_core_client: OSCoreClient,
        cache_name="default",
        cache_period_seconds=120,
    ):
        self.os_core_client = os_core_client

        if cache_name is not None:
            from django.core.cache import caches  # type: ignore

            self.cache = caches[cache_name] if cache_name else None
            self.cache_period_seconds = cache_period_seconds
        else:
            self.cache = None
            self.cache_period_seconds = None

    def _cache_key(
        self, person_id: str, ref: Union[str, int], ref_type: Optional[RefType] = None
    ) -> str:
        """ref can be either an org_ref or a legacy int ref here"""
        if ref_type:
            return f"{self._cache_prefix}:{person_id}:{ref}:{ref_type}"
        return f"{self._cache_prefix}:{person_id}:{ref}"

    def _global_cache_key(self, person_id: str) -> str:
        return f"{self._cache_prefix}:{person_id}:global"

    def _cache_read(
        self, person_id: str, org_ref: str, ref_type: Optional[RefType] = None
    ) -> Optional[CachedPermissionsDoc]:
        if not self.cache:
            return None

        keys = [self._global_cache_key(person_id)]
        if org_ref != "global":
            keys.append(self._cache_key(person_id, org_ref, ref_type))

        try:
            fetched = self.cache.get_many(keys)
        except Exception as e:
            raise ServerError("Error reading from cache") from e

        if len(fetched.keys()) != len(keys):
            logger.debug(f"Not all keys found in cache, got: {fetched}")
            return None

        logger.debug(f"Successful cache read: {fetched}")

        return fetched

    def _cache_transform(
        self, person_id: str, permissions_doc: FetchedPermissionsDoc
    ) -> CachedPermissionsDoc:
        logger.debug(f"Transforming to cache representation: {permissions_doc}")

        cache_doc = {}
        for ref, scopes in permissions_doc.items():
            org_ref: str
            ref_type: Optional[RefType] = None

            if ":" in ref:
                split = ref.split(":")
                ref_type, org_ref = cast(RefType, split[0]), split[1]
            else:
                org_ref = ref

            joined_scopes = f"{'|'.join(scopes)}|"
            cache_doc[
                self._cache_key(person_id, org_ref, ref_type)
            ] = joined_scopes

        return cache_doc

    def _cache_write(self, doc: CachedPermissionsDoc) -> None:
        if self.cache:
            logger.debug(f"Writing to cache: {doc}")
            try:
                self.cache.set_many(doc, timeout=self.cache_period_seconds)
            except Exception as e:
                raise ServerError("Error writing to cache") from e

    def _has_permission(
        self,
        person_id: UUIDType,
        scope: str,
        org_ref: Union[UUIDType, Literal["global"], int],
        ref_type: Optional[RefType] = None,
    ) -> bool:
        person_id = str(person_id)
        org_ref = str(org_ref)

        cached_doc = self._cache_read(person_id, org_ref, ref_type)
        if not cached_doc:
            if ref_type is not None:
                fetched_doc = self.os_core_client.fetch_permissions_for_location(
                    person_id, int(org_ref), ref_type
                )
            else:
                fetched_doc = self.os_core_client.fetch_permissions(person_id, org_ref)
            cached_doc = self._cache_transform(person_id, fetched_doc)
            self._cache_write(cached_doc)

        if cached_doc:
            for scopes in cached_doc.values():
                if f"{scope}|" in scopes:
                    return True

        return False

    def has_global_permission(self, person_id: UUIDType, scope: str) -> bool:
        return self._has_permission(person_id, scope, "global")

    @overload  # noqa: F811
    def has_permission(
        self, person_id: UUIDType, scope: str, org_ref: UUIDType
    ) -> bool:
        ...

    @overload  # noqa: F811
    def has_permission(
        self, person_id: UUIDType, scope: str, org_ref: int, ref_type: RefType
    ) -> bool:
        ...

    def has_permission(  # noqa: F811
        self,
        person_id: UUIDType,
        scope: str,
        org_ref: Union[UUIDType, int],
        ref_type: Optional[RefType] = None,
    ) -> bool:
        return self._has_permission(person_id, scope, org_ref, ref_type)



import codecs
import os

import setuptools


here = os.path.abspath(os.path.dirname(__file__))


with codecs.open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()

about = {}
with codecs.open(os.path.join(here, 'mbq', 'client', '__version__.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)

setuptools.setup(
    name=about['__title__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/x-rst',
    version=about['__version__'],
    license=about['__license__'],
    url=about['__url__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    maintainer=about['__author__'],
    maintainer_email=about['__author_email__'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
    ],
    install_requires=[
        'requests>=2.21.0,<3.0.0',
        'typing_extensions>=3.7.2',
    ],
    keywords='token access authorization',
    packages=setuptools.find_packages(),
    zip_safe=True,
)





class Authenticator:

    def __init__(self, service_name, token_manager):
        self._token_manager = token_manager
        self.service_name = service_name

    def __call__(self, request):
        request.headers['Authorization'] = 'Bearer {}'.format(
            self._token_manager.get_token(service_name=self.service_name)
        )
        return request
